//# Parset.h: class/struct that holds the Parset information
//# Copyright (C) 2008-2017  ASTRON (Netherlands Institute for Radio Astronomy)
//# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
//#
//# This file is part of the LOFAR software suite.
//# The LOFAR software suite is free software: you can redistribute it and/or
//# modify it under the terms of the GNU General Public License as published
//# by the Free Software Foundation, either version 3 of the License, or
//# (at your option) any later version.
//#
//# The LOFAR software suite is distributed in the hope that it will be useful,
//# but WITHOUT ANY WARRANTY; without even the implied warranty of
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//# GNU General Public License for more details.
//#
//# You should have received a copy of the GNU General Public License along
//# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
//#
//# $Id$

#ifndef LOFAR_INTERFACE_PARSET_H
#define LOFAR_INTERFACE_PARSET_H

// \file
// class/struct that holds the Parset information

//# Never #include <config.h> or #include <lofar_config.h> in a header file!

#include <string>
#include <vector>
#include <numeric>
#include <sstream>
#include <ostream>
#include <algorithm>

#include <Common/ParameterSet.h>
#include <Common/StringUtil.h>
#include <Common/StreamUtil.h>
#include <Stream/Stream.h>
#include <CoInterface/OutputTypes.h>
#include <CoInterface/MultiDimArray.h>


namespace LOFAR
{
  namespace Cobalt
  {

    enum StokesType { STOKES_I = 0, STOKES_IQUV, STOKES_XXYY, INVALID_STOKES = -1 };

    StokesType stokesType( const std::string &name );
    size_t nrStokes( StokesType type );
    std::string stokesType( StokesType type );

    // All settings relevant for an observation (well, it should become that,
    // we don't copy all Parset values yet!).
 
    struct ObservationSettings {
      /*
       * Generic information
       */

      // Whether the observation runs at real time. Non-real time
      // observations are not allowed to lose data.
      //
      // key: Cobalt.realTime
      bool realTime;

      // The SAS/MAC observation number
      //
      // key: Observation.ObsID
      unsigned observationID;

      // The MoM observation number
      //
      // key: Observation.momID
      unsigned momID;

      // Command stream, or null: if not used
      //
      // key: Cobalt.commandStream
      std::string commandStream;

      // Specified observation start time, in seconds since 1970.
      //
      // key: Observation.startTime
      double startTime;

      // Specified observation stop time, in seconds since 1970.
      //
      // key: Observation.stopTime
      double stopTime;

      // The station clock, in MHz (200 or 160)
      //
      // key: Observation.sampleClock
      unsigned clockMHz;

      // The station clock, in Hz
      unsigned clockHz() const;

      // The bandwidth of a single subband, in Hz
      double subbandWidth() const;

      // The length of an input sample, in s
      double sampleDuration() const;

      // The number of samples in one block of one subband.
      //
      // key: Cobalt.blockSize
      size_t blockSize;

      size_t nrBlocks() const;

      size_t nrRspRawBlocks() const;

      // The number of seconds represented by each block.
      double blockDuration() const;

      // The number of bits in each input sample (16, 8, or 4)
      //
      // key: Observation.nrBitsPerSample
      unsigned nrBitsPerSample;

      // The number of polarisations. Set to 2.
      unsigned nrPolarisations;

      // The number of cross polarisations.
      unsigned nrCrossPolarisations() const;

      struct Corrections {
        // Whether the station band pass should be corrected for
        //
        // key: Cobalt.correctBandPass
        bool bandPass;

        // Whether the station clock offsets should be corrected for
        //
        // key: Cobalt.correctClocks
        bool clock;

        // Whether to dedisperse tied-array beams
        //
        // key: Cobalt.BeamFormer.coherentDedisperseChannels
        bool dedisperse;
      };
      
      struct Corrections corrections;

      struct DelayCompensation {
        // Whether geometric delays should be compensated for
        //
        // key: Cobalt.delayCompensation
        bool enabled;

        // The ITRF position to compensate delays to
        //
        // key: Observation.referencePhaseCenter
        std::vector<double> referencePhaseCenter;
      };
      
      struct DelayCompensation delayCompensation;

      /*
       * Station / Antenna field information
       */

      // The selected antenna set (LBA, HBA_DUAL, HBA_ZERO, etc)
      //
      // key: Observation.antennaSet
      std::string antennaSet;

      // The selected band filter (LBA_30_70, etc)
      //
      // key: Observation.bandFilter
      std::string bandFilter;

      struct AntennaFieldName {
        std::string station;
        std::string antennaField;

        AntennaFieldName(const std::string &station, const std::string &antennaField)
        : station(station),
          antennaField(antennaField)
        { }

        std::string fullName() const {
          return station + antennaField;
        }

        bool operator==(const AntennaFieldName &other) const {
          return station == other.station &&
                 antennaField == other.antennaField;
        }

        bool operator==(const std::string &other) const {
          return fullName() == other;
        }

        // Returns the indices in fullSet for each element of subSet.
        static std::vector<unsigned> indices(const std::vector<AntennaFieldName> &subSet,
                                             const std::vector<AntennaFieldName> &fullSet);

        // Returns a list of the station names.
        static std::vector<std::string> names( const std::vector<AntennaFieldName> &list );
      };

      struct AntennaField {
        // The name of the antenna field (CS001LBA, etc)
        //
        // key: OLAP.storageStationNames[antennaFieldIdx]
        std::string name;

        // The input streams descriptors
        //
        // key: PIC.Core.CS001LBA.RSP.ports
        std::vector<std::string> inputStreams;

        // The node name on which this antenna field is received
        //
        // key: PIC.Core.CS001LBA.RSP.receiver
        std::string receiver;

        // Correction on the station clock, in seconds
        //
        // key: PIC.Core.CS001LBA.clockCorrectionTime
        double clockCorrection;

        // The phase center for which the antenna field beams are corrected, in
        // ITRF [x,y,z].
        //
        // key: PIC.Core.CS001LBA.phaseCenter
        std::vector<double> phaseCenter;

        // The phase correction for this antenna field, in radians.
        //
        // key: PIC.Core.CS001.LBA_INNER.LBA_30_70.phase0.X
        // key: PIC.Core.CS001.LBA_INNER.LBA_30_70.phase0.Y
        struct {
          double x;
          double y;
        } phase0;

        // The delay correction for this station, in seconds
        //
        // key: PIC.Core.CS001.LBA_INNER.LBA_30_70.delay.X
        // key: PIC.Core.CS001.LBA_INNER.LBA_30_70.delay.Y
        struct {
          double x;
          double y;
        } delay;


        // The RSP board to which each subband is mapped
        //
        // key: Observation.Dataslots.CS001LBA.RSPBoardList
        //  or: Observation.rspBoardList
        std::vector<unsigned> rspBoardMap; // [subband]

        // The RSP slot to which each subband is mapped
        //
        // key: Observation.Dataslots.CS001LBA.DataslotList
        //  or: Observation.rspSlotList
        std::vector<unsigned> rspSlotMap;  // [subband]

        // Allow comparing by name, useful for std::find.
        bool operator==(const std::string &other) const { return name == other; }
        bool operator==(const AntennaFieldName &other) const { return name == other.fullName(); }
      };

      // All antenna fields specified as input
      //
      // length: len(OLAP.storageStationNames)
      std::vector<struct AntennaField> antennaFields;
      std::vector<struct AntennaFieldName> antennaFieldNames;

      // A list of the stations used in the observation
      //
      // key: Observation.VirtualInstrument.stationList
      std::string rawStationList;

      ssize_t antennaFieldIndex(const std::string &name) const;

      /*
       * Resources information:
       *   - what hardware we use (cpus/gpus)
       */ 

      struct Node {
        // MPI rank of this node, is the
        // same as the index in the `nodes' vector.
        int rank;

        // (Symbolic) name
        std::string name;

        // Host name
        std::string hostName;

        // CPU socket number to bind to, or -1 to not bind to any
        int cpu;

        // CUDA GPU numbers to bind to
        std::vector<unsigned> gpus;

        // NIC(s) to bind to (comma seperated)
        //
        // E.g. "mlx4_0", "mlx4_1", "eth0", etc
        std::string mpi_nic; // for MPI
        std::string out_nic; // to outputProc
      };

      std::vector<struct Node> nodes;

      // Cluster where data writers run
      // TODO: move to per output type and read Output_XXX.storageClusterName keys
      std::string outputCluster;

      /*
       * Spectral resolution information
       */

      struct Subband {
        // Index (e.g. 0..243)
        //
        // set to: equals the index in the subbands vector
        unsigned idx;


        // Index of this subband in the SAP it is part of
        //
        // Calculated based on  Observation.Beam[x].subbandList
        unsigned idxInSAP;

        // Index at station (e.g. 100..343)
        //
        // key: Observation.subbandList[idx]
        unsigned stationIdx;

        // SAP number
        //
        // key: Observation.beamList[idx]
        unsigned SAP;

        // Central frequency (Hz)
        //
        // set to: subbandWidth() * (512 * (nyquistZone() - 1) + stationIdx)
        double centralFrequency;
      };

      // The list of subbands
      //
      // length: len(Observation.subbandList)
      std::vector<struct Subband> subbands;

      /*
       * Pointing information
       */
      struct Direction {
        // Coordinate type (J2000, etc)
        //
        // key: *.directionType
        std::string type;

        // Two angles within the coordinate type (RA/DEC, etc)
        //
        // key: *.angle1
        // key: *.angle2
        double angle1;
        double angle2;
      };

      struct SAP {
        // Direction in which the SAP points
        //
        // key: Observation.Beam[sapIdx].*
        struct Direction direction;

        // The list of subbands in this SAP
        //
        // key: Observation.Beam[idx].subbandList 
        std::vector<struct Subband> subbands;

        // Name of target
        //
        // key: Observation.Beam[sapIdx].target
        std::string target;

        // Return the list of indices of our subbands
        // within the global settings.subbands list.
        vector<unsigned> subbandIndices() const;
      };

      // All station beams
      //
      // length: Observation.nrBeams
      std::vector<struct SAP> SAPs;

      struct AnaBeam {
        // Whether the observation employs an analog beam
        //
        // key: Observation.antennaSet starts with "HBA"
        bool enabled;

        // Direction in which the analog beam points
        //
        // key: Observation.AnaBeam[0].*
        struct Direction direction;
      };

      // The analog beam, if any
      struct AnaBeam anaBeam;

      struct FileLocation {
        string cluster;
        string host;
        string directory;
        string filename;

        // Make FileLocation objects comparable and equality testable,
        // so we can prune dups in a vector using sort + unique.
        bool operator<(const FileLocation& rhs) const {
          int cmp = cluster.compare(rhs.cluster);
          if (cmp < 0) return true;
          if (cmp > 0) return false;
          cmp = host.compare(rhs.host);
          if (cmp < 0) return true;
          if (cmp > 0) return false;
          cmp = directory.compare(rhs.directory);
          if (cmp < 0) return true;
          if (cmp > 0) return false;
          cmp = filename.compare(rhs.filename);
          if (cmp < 0) return true;
          return false;
        }

        bool operator==(const FileLocation& rhs) const {
          return filename  == rhs.filename  && host    == rhs.host &&
                 directory == rhs.directory && cluster == rhs.cluster;
        }

        std::string toString() const {
          std::string s;
          if (!cluster.empty()) {
            s = cluster + ':';
          }
          return s + host + ':' + directory + '/' + filename;
        }
      };

      /* ===============================
       * Correlator pipeline information
       * ===============================
       */

      struct Correlator {
        // Whether to output correlated data
        //
        // key: Observation.DataProducts.Output_Correlated.enabled
        bool enabled;

        // Number of requested frequency channels per subband
        //
        // key: Cobalt.Correlator.nrChannelsPerSubband
        unsigned nrChannels;

        // The bandwidth of a single channel, in Hz
        //
        // set to: subbandWidth() / nrChannels
        double channelWidth;

        // The number of samples in one block of one channel.
        //
        // key: OLAP.CNProc.integrationSteps
        size_t nrSamplesPerBlock;

        // The number of blocks to integrate to obtain the final
        // integration time.
        //
        // If >1, the integration time is longer than the blockSize.
        //
        // key: Cobalt.Correlator.nrBlocksPerIntegration
        size_t nrBlocksPerIntegration;

        // The number of subblocks to produce per block.
        //
        // If >1, the integration time is shorter than the blockSize.
        //
        // key: Cobalt.Correlator.nrIntegrationsPerBlock
        size_t nrIntegrationsPerBlock;

        // The number of integrations that will be emitted between
        // the start and end time of the observation. This is the
        // expected number of CorrelatedData blocks written in the
        // MeasurementSet.
        size_t nrIntegrations;

        // The number of samples to integrate over.
        size_t nrSamplesPerIntegration() const;

        // The total integration time of all blocks, in seconds.
        double integrationTime() const;

        struct Station {
          // The name of this (super)station
          //
          // key: OLAP.tiedArrayStationNames
          std::string name;

          // The list of (input) stations indices to sum for this station
          //
          // key: OLAP.CNProc.tabList
          std::vector<size_t> inputStations;
        };

        // Super-station beam former. NOTE: Not yet supported!! Only
        // here to register how the parset keys work. Until then, this
        // array is the same as the (input) stations array above.
        //
        // The aim is to output this list of stations instead of the
        // ones used for input.
        std::vector<struct Station> stations;

        struct File {
          size_t streamNr;

          struct FileLocation location;

          // Where to create the (meta)data, before moving them to their
          // final location. If empty, directly create the (meta)data in
          // their final location.

          // data is moved through a single staging dir, which needs to
          // be on the same mount point as the final location.
          //
          // metaData is moved through 2 staging dirs. The files are
          // created in stagingDir1, moved to stagingDir2 to move to
          // the target mount point, and then renamed to their final location.
          std::string dataStagingDirTargetFS; // stage metadata here on target FS
          std::string metaDataStagingDirCreateFS; // create metadata here
          std::string metaDataStagingDirTargetFS; // stage metadata here on target FS
        };

        // The list of files to write, indexed by subband
        std::vector<struct File> files; // [subband]
      };

      struct Correlator correlator;

      /* ===============================
       * Beamformer pipeline information
       * ===============================
       */

      struct BeamFormer {
        // Whether beam forming was requested.
        //
        // key: Observation.DataProducts.Output_Beamformed.enabled
        bool enabled;

        struct File {
          size_t streamNr;

          size_t sapNr;
          size_t tabNr;
          size_t stokesNr;
          size_t partNr;
          bool coherent;

          // this TAB is the ....th coherent TAB in this SAP
          size_t coherentIdxInSAP;

          // this TAB is the ....th incoherent TAB in this SAP
          size_t incoherentIdxInSAP;

          struct FileLocation location;

          // this file stores [firstSubbandIdx, lastSubbandIdx)
          // interpretation is same as in globalSubbandIdx, i.e. [0, 488)
          unsigned firstSubbandIdx;
          unsigned lastSubbandIdx; // exclusive
        };

        // The list of files to write, one file
        // per part/stokes.
        std::vector<struct File> files;

        // Number of channels per subband for delay compensation.
        // Equal to the size of the first FFT. Power of two.
        unsigned nrDelayCompensationChannels;

        // Number of channels per subband for bandpass correction, narrow band
        // flagging, beamforming, and coherent dedispersion.
        // Power of two and at least nrDelayCompensationChannels.
        unsigned nrHighResolutionChannels;

        // Are we in fly's eye mode?
        bool doFlysEye;

        struct TAB {
          // The (absolute) direction where the TAB points to.
          //
          // key: Observation.Beam[sap].TiedArrayBeam[tab].*
          struct Direction direction;

          // Whether the beam is coherent (or incoherent)
          //
          // key: Observation.Beam[sap].TiedArrayBeam[tab].coherent
          bool coherent;

          // The DM with which to dedisperse this beam, or
          // 0.0 for no dedispersion.
          //
          // key: Observation.Beam[sap].TiedArrayBeam[tab].dispersionMeasure
          double dispersionMeasure;

          // The list of files to write, one file
          // per part/stokes.
          std::vector<struct File> files; 
        };

        struct SAP {
          // The TABs to form in this SAP
          //
          // size: Observation.Beam[sap].nrTiedArrayBeams
          std::vector<struct TAB> TABs;

          // calculated at construction time
          size_t nrCoherent;
          size_t nrIncoherent;

          // list of subbands in this sap
          vector<unsigned> subbandIndices;
        };

        // Antenna fields to use for beam forming.
        // Must be a subset of the observation antenna-field list
        //
        // Derived from key: Cobalt.Beamformer.stationList
        // TODO: split up in CoherentStokes and IncoherentStokes antenna field name lists
        std::vector<AntennaFieldName> antennaFieldNames;

        // All SAPs, with information about the TABs to form.
        //
        // size: len(Observation.nrBeams)
        std::vector<struct SAP> SAPs;

        // Return whether there are any (in)coherent TABs specified in the
        // observation. These functions are valid even if enabled == false.
        bool anyCoherentTABs() const;
        bool anyIncoherentTABs() const;

        size_t maxNrTABsPerSAP() const;
        size_t maxNrCoherentTABsPerSAP() const;
        size_t maxNrIncoherentTABsPerSAP() const;

        struct StokesSettings {
          // Reflection: whether this struct captures
          // coherent or incoherent stokes settings.
          bool coherent;

          // The type of stokes to output
          //
          // key: *.which
          StokesType type;

          // The number of stokes to output
          //
          // set to: nrStokes(type)
          unsigned nrStokes;

          // The requested number of channels
          //
          // key: *.channelsPerSubband
          unsigned nrChannels;

          // The number of samples that need
          // to be integrated temporally, per channel.
          //
          // key: *.timeIntegrationFactor
          size_t timeIntegrationFactor;

          // The number of samples per channel
          size_t nrSamples;

          // The number of subbands to store in each file.
          // The last file can have fewer subbands.
          //
          // key: *.subbandsPerFile
          size_t nrSubbandsPerFile;
        };

        // Settings for Coherent Stokes output
        //
        // key: Cobalt.BeamFormer.CoherentStokes.*
        struct StokesSettings coherentSettings;

        // Settings for Incoherent Stokes output
        //
        // key: Cobalt.BeamFormer.IncoherentStokes.*
        struct StokesSettings incoherentSettings;


        // Size of FFT for coherent dedispersion
        //
        // key: Cobalt.BeamFormer.dedispersionFFTsize
        size_t dedispersionFFTsize;
      };

      struct BeamFormer beamFormer;

      /* ===============================
       * RSPRaw collection information
       * ===============================
       */

      struct RSPRaw {
          // Whether RSP raw data collection was requested.
          //
          // key: Observation.DataProducts.Output_RSPRaw.enabled
          // Default: false
          bool enabled;

          // Specified RSP raw data dump / piggy backing start time
          // in seconds since 1970.
          //
          // key: Cobalt.RSPRaw.startTime
          // Default: value of Observation.startTime
          time_t startTime;

          // Specified raw RSP data dump / piggy backing stop time
          // in seconds since 1970.
          //
          // key: Cobalt.RSPRaw.stopTime
          // Default: value of Observation.stopTime
          time_t stopTime;

          // All antenna fields specified for RSP raw output
          // Must be a subset of the observation antenna-field list
          //
          // Derived from key: Cobalt.RSPRaw.stationList
          // Default value: antennaFieldNames (observation wide)
          std::vector<AntennaFieldName> antennaFieldNames;

          // A vector with per RSP board (for all stations) the number of beamlets,
          // to send out from each station RSP frame.
          // Vector size must be the (max) number of antenna field input streams (typically 4).
          // (This is a cheap variant of full (beam, subband) selection per (station, board).)
          //
          // key: Cobalt.RSPRaw.nrBeamletsPerBoardList
          // An absent key, missing values (vector too small), or too high value means all.
          std::vector<unsigned> nrBeamletsPerBoardList;

          struct File {
              // Index into *RSP raw* antennaFieldNames
              unsigned antennaFieldNameIdx;

              // Stream number, aka sender RSP board nr (typically 0-3)
              unsigned streamNr;

              struct FileLocation location;
          };

          // The list of files to write
          std::vector<struct File> files;
      };

      struct RSPRaw rspRaw;

      // Returns the Nyquist zone number based on bandFilter.
      unsigned nyquistZone() const;

      // Constructs the antenna fields ("CS001", "HBA0") etc from a set of stations
      // ("CS001", "CS002") and the antenna set.
      static std::vector<struct AntennaFieldName>
      expandAntennaFieldNames(const std::vector<std::string> &stations,
                        const std::string &antennaSet);

      // List of host names to start outputProc on
      std::vector<std::string> outputProcHosts;
    }; // struct ObservationSettings

    // Reads a ParameterSet from a Stream
    void readParameterSet(Stream &, ParameterSet &);


    // The Parset class is a public struct that can be used as base-class
    // for holding Parset related information.
    // It can be instantiated with a parset containing Parset information.
    class Parset : public ParameterSet
    {
    public:
      Parset();
      Parset(const std::string &name);
      
      // Read a parset from a Stream
      Parset(Stream *);

      // Transform the parset into an ObservationSettings object
      struct ObservationSettings observationSettings() const;

      // Fill the settings based on the ParameterSet keys.
      // Call this if keys are added or changed.
      void updateSettings();

      struct ObservationSettings settings;

      std::string                 name() const;
      void                        check() const;

      void                        write(Stream *) const;

      double                      getRealStopTime() const;
      unsigned                    nrTabStations() const;
      unsigned                    nrMergedStations() const;
      std::vector<std::string>    mergedStationNames() const;
      unsigned                    nrBaselines() const;
      double                      sampleDuration() const;
      unsigned                    nrBitsPerSample() const;
      size_t                      nrBytesPerComplexSample() const;
      std::vector<double>         position(const string &name) const;
      MultiDimArray<double,2>     positions() const;
      std::string                 positionType() const;
      unsigned                    dedispersionFFTsize() const;

      unsigned                    nrProcessedOutputTypes() const;
      bool                        outputThisType(OutputType) const;

      unsigned nrStreams(OutputType, bool force = false) const;
      std::string getHostName(OutputType, unsigned streamNr) const;
      std::string getFileName(OutputType, unsigned streamNr) const;
      std::string getDirectoryName(OutputType, unsigned streamNr) const;
      unsigned getRSPRawOutputStreamIdx(const std::string& antennaFieldName, unsigned boardNr) const;

      double channel0Frequency( size_t subband, size_t nrChannels ) const;

      std::vector<double>         itsStPositions;

      std::string                 PVSS_TempObsName() const;

    private:
      const std::string itsName;

      mutable std::string itsWriteCache;

      void                        checkInputConsistency() const;

      time_t                      getTime(const std::string &name, const std::string &defaultValue) const;

      std::vector<double>         centroidPos(const string &stations) const;

      std::vector<struct ObservationSettings::AntennaFieldName>
                                  getOutputTypeAntennaFieldNames(const std::string& stationListKey,
                                                                 const std::vector<std::string>& stations,
                                                                 const std::string& antennaSet) const;

      std::vector<struct ObservationSettings::FileLocation>
                                  getFileLocations(const std::string outputType) const;

      // Returns whether nodeName has to participate in the observation
      // given antenna fields, antenna mode, and configured antenna field streams.
      // The nodeName is e.g. "cbt001_0", or "gpu01_0", or "localhost".
      bool                        nodeReadsAntennaFieldData(const struct ObservationSettings& settings,
                                                            const std::string& nodeName) const;

      static double               distanceVec3(const std::vector<double>& pos,
                                               const std::vector<double>& ref);
      static double               maxDelayDistance(const struct ObservationSettings& settings);
      static double               maxObservationFrequency(const struct ObservationSettings& settings,
                                                          double subbandWidth);
      static unsigned             calcNrDelayCompensationChannels(const struct ObservationSettings& settings);
    };

    std::ostream &operator<<(std::ostream &os, const ObservationSettings::AntennaFieldName &fieldName);
  } // namespace Cobalt
} // namespace LOFAR

#endif

