//# TABTranspose.cc
//# Copyright (C) 2012-2013  ASTRON (Netherlands Institute for Radio Astronomy)
//# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
//#
//# This file is part of the LOFAR software suite.
//# The LOFAR software suite is free software: you can redistribute it and/or
//# modify it under the terms of the GNU General Public License as published
//# by the Free Software Foundation, either version 3 of the License, or
//# (at your option) any later version.
//#
//# The LOFAR software suite is distributed in the hope that it will be useful,
//# but WITHOUT ANY WARRANTY; without even the implied warranty of
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//# GNU General Public License for more details.
//#
//# You should have received a copy of the GNU General Public License along
//# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
//#
//# $Id$

#include <lofar_config.h>

#include "TABTranspose.h"

#include <algorithm>
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>

#include <Common/LofarLogger.h>
#include <Common/Timer.h>
#include <ApplCommon/PVSSDatapointDefs.h>
#include <CoInterface/TimeFuncs.h>
#include <CoInterface/BudgetTimer.h>

using namespace std;
using boost::format;
using boost::lexical_cast;

namespace LOFAR {
namespace Cobalt {
namespace TABTranspose {

// A Subband here is the part of a Block (see below) that contains the data
// values in a subband during a block duration (e.g. 1 second).
Subband::Subband( size_t nrSamples, size_t nrChannels )
: 
  data(boost::extents[nrSamples][nrChannels])
{
  id.fileIdx = 0;
  id.subband = 0;
  id.block   = 0;
}


// Used by a MultiSender in rtcp to write a Subband through a Stream to a
// Receiver in outputProc. 
void Subband::write(Stream &stream) const {
  stream.write(&id, sizeof id);

  size_t dim1 = data.shape()[0];
  size_t dim2 = data.shape()[1];
  stream.write(&dim1, sizeof dim1);
  stream.write(&dim2, sizeof dim2);
  stream.write(data.origin(), data.num_elements() * sizeof *data.origin());

  LOG_DEBUG_STR("Written block " << id);
}


// Used by a Receiver in outputProc to read a Subband through a Stream from a
// MultiSender in rtcp.
void Subband::read(Stream &stream) {
  LOG_DEBUG_STR("Reading block id");
  stream.read(&id, sizeof id);
  LOG_DEBUG_STR("Read block id " << id);

  size_t dim1, dim2;
  stream.read(&dim1, sizeof dim1);
  stream.read(&dim2, sizeof dim2);
  data.resize(boost::extents[dim1][dim2]);
  stream.read(data.origin(), data.num_elements() * sizeof *data.origin());
  LOG_DEBUG_STR("Read block " << id);
}


std::ostream &operator<<(std::ostream &str, const Subband::BlockID &id)
{
  return str << "stream " << id.fileIdx << " block " << id.block << " subband " << id.subband;
}


// A Block here is a block duration (e.g. 1 sec) of data with all the Subbands
// that should go info a single file (part).
Block::Block( size_t fileIdx, size_t blockIdx, size_t nrSubbands, size_t nrSamples, size_t nrChannels )
:
  fileIdx(fileIdx),
  blockIdx(blockIdx),
  nrSamples(nrSamples),
  nrSubbands(nrSubbands),
  nrChannels(nrChannels),
  subbandCache(nrSubbands, NULL),
  nrSubbandsLeft(nrSubbands)
{
}


// Used by the BlockCollector to add a received Subband to a Block.
// Runs in outputProc.
void Block::addSubband( SmartPtr<Subband> &subband ) {
  ASSERT(nrSubbandsLeft > 0);

  const Subband::BlockID id = subband->id;

  // Only add subbands that match our ID
  ASSERTSTR(id.fileIdx == fileIdx,  "Got fileIdx " << id.fileIdx << ", expected " << fileIdx);
  ASSERTSTR(id.subband < nrSubbands, "Got subband " << id.subband << ", expected < " << nrSubbands);
  ASSERTSTR(id.block   == blockIdx, "Got block " << id.block << ", expected " << blockIdx);

  ASSERT(id.subband < nrSubbands);
  ASSERT(subband->data.shape()[0] == nrSamples);
  ASSERT(subband->data.shape()[1] == nrChannels);

  // Subbands should not arrive twice
  ASSERT(subbandCache[id.subband] == NULL);

  subbandCache[id.subband] = subband;
  ASSERT(subband == NULL);

  nrSubbandsLeft--;

  LOG_DEBUG_STR("Block: adding " << id << ", " << nrSubbandsLeft << " subbands left");
}


void Block::write( BeamformedData &output ) {
  // Check dimensions
  ASSERT( output.samples.shape()[0] == nrSamples );
  ASSERT( output.samples.shape()[1] == nrSubbands );
  ASSERT( output.samples.shape()[2] == nrChannels );

  // Set annotation
  output.setSequenceNumber(blockIdx);

  // Set data

  /*
   * Input:
   *   the set of subbandCache[nrSubbands],
   *   each of which is either NULL, 
   *   or (indirectly) a pointer to [nrSamples][nrChannels]
   * Output:
   *   output.samples[nrSamples][nrSubbands][nrChannels]
   *
   * Which means we're transposing nrChannels * sizeof(float) bytes at a time,
   * which is especially painful if nrChannels == 1.
   *
   * The implemented strategy coalesces the writes in batches. Large batches
   * will cause TLB misses due to the larger number of arrays we need to index.
   */

  // On CEP2, optimal is MAXBATCHSIZE=4 and using 4 threads (in the for loop)
  // On CEP4, optimal is MAXBATCHSIZE=488 and using 1 thread.
  //
  // We want real-time performance, (see tTABTranspose | grep write, should produce
  // numbers <1s), but if we do not need multiple threads, doing the work in one
  // is more efficient.
  const size_t MAXBATCHSIZE = 488;

  // Stride between samples in output
  const ptrdiff_t dst_sample_stride = output.samples.strides()[0];
  // Stride between subbands in output
  const ptrdiff_t dst_sb_stride = output.samples.strides()[1];

  const vector<float> zeroes(nrChannels * nrSamples, 0.0f);

  for (size_t subbandBase = 0; subbandBase < subbandCache.size(); subbandBase += MAXBATCHSIZE) {
    // Determine actual batch size
    const size_t BATCHSIZE = std::min(MAXBATCHSIZE, subbandCache.size() - subbandBase);

    // Collect source pointers for all our subbands, or to our zeroes array if
    // a subband is not available. Each element points to an array of dimensions
    //   [nrSamples][nrChannels]
    const float *src[MAXBATCHSIZE];

    for (size_t i = 0; i < BATCHSIZE; ++i) {
      size_t subbandIdx = subbandBase + i;

      src[i] = subbandCache[subbandIdx] ? subbandCache[subbandIdx]->data.origin() : &zeroes[0];
    }

    // Pointer to walk over all samples
    float *dst = &output.samples[0][subbandBase][0];

    if (nrChannels == 1) {
      /* Use assignment to copy data */
      for (size_t t = 0; t < nrSamples; ++t) {
        float *sample = dst;
        dst += dst_sample_stride;

        for (size_t i = 0; i < BATCHSIZE; ++i) {
          *sample = src[i][t];

          sample += dst_sb_stride;
        }
      }
    } else {
      /* Use memcpy to copy data */
      for (size_t t = 0; t < nrSamples; ++t) {
        float *sample = dst;
        dst += dst_sample_stride;

        for (size_t i = 0; i < BATCHSIZE; ++i) {
          memcpy(sample, &src[i][t * nrChannels], nrChannels * sizeof(float));

          sample += dst_sb_stride;
        }
      }
    }
  }

  // Update flags
  for (size_t sb = 0; sb < subbandCache.size(); sb++) {
      output.flags[sb].reset(); 

      if (!subbandCache[sb])
          output.flags[sb].include(0, nrSamples);
  }

  // Report summary
  if (complete())
    LOG_DEBUG_STR("[block " << blockIdx << " stream " << fileIdx << "] [Block] Written " << (nrSubbands - nrSubbandsLeft) << " subbands, lost " << nrSubbandsLeft << " subbands.");
  else
    LOG_INFO_STR("[block " << blockIdx << " stream " << fileIdx << "] [Block] Written " << (nrSubbands - nrSubbandsLeft) << " subbands, lost " << nrSubbandsLeft << " subbands.");
}


bool Block::complete() const {
  return nrSubbandsLeft == 0;
}


// The BlockCollector collects blocks from different rtcp processes for a TAB.
// More precisely, we have one BlockCollector per file (i.e. part).
BlockCollector::BlockCollector( Pool<BeamformedData> &outputPool, size_t fileIdx, size_t nrSubbands, size_t nrChannels, size_t nrSamples, size_t nrBlocks, size_t maxBlocksInFlight, const std::string &logPrefix, double blockDuration )
:
  logPrefix(logPrefix + "[BlockCollector] "),

  nrFullBlocksLost(0),
  nrBlockSubbandsLost(0),
  nrBlockSubbandsReceived(0),
  nrBlockSubbandsTooLate(0),

  // drop = false: we drop at the output, not at the input, but we do want to protect against unbounded growth
  inputQueue(str(format("BlockCollector::inputQueue [stream %u]") % fileIdx), (1 + maxBlocksInFlight) * nrSubbands, false),
  outputQueue(str(format("BlockCollector::outputQueue [stream %u]") % fileIdx), 3, false),

  outputPool(outputPool),

  fileIdx(fileIdx),
  nrBlocks(nrBlocks),
  nrSubbands(nrSubbands),
  nrChannels(nrChannels),
  nrSamples(nrSamples),

  blockDuration(blockDuration),

  maxBlocksInFlight(maxBlocksInFlight),
  canDrop(maxBlocksInFlight > 0),
  lastEmitted(-1),

  inputThread(this, &BlockCollector::inputLoop, str(format("BC:input %u") % fileIdx)),
  outputThread(this, &BlockCollector::outputLoop, str(format("BC:output %u") % fileIdx))
{
  ASSERT(nrSubbands > 0);
  ASSERT(nrChannels > 0);
  ASSERT(nrSamples > 0);
}


BlockCollector::~BlockCollector()
{
  // Make SURE the threads can finish, regardless of whether finish() was called
  inputQueue.noMore();
  outputQueue.noMore();
}


void BlockCollector::addSubband( SmartPtr<Subband> &subband ) {
  inputQueue.append(subband);
}


void BlockCollector::inputLoop() {
  SmartPtr<Subband> subband;

  BudgetTimer processTimer(
    "BlockCollector::inputLoop: weaving subbands",
    blockDuration / nrSubbands,
    true, true);

  while ((subband = inputQueue.remove()) != NULL) {
    BudgetTimer::StartStop ss(processTimer);

    processSubband(subband);
  }
}


void BlockCollector::outputLoop() {
  SmartPtr<Block> block;

  BudgetTimer writeTimer(
    "BlockCollector::outputLoop: data transpose/zeroing",
    blockDuration,
    true, true);

  while ((block = outputQueue.remove()) != NULL) {
    SmartPtr<BeamformedData> output = outputPool.free.remove();

    {
      BudgetTimer::StartStop ss(writeTimer);
      block->write(*output);
    }

    outputPool.filled.append(output);
  }

  outputPool.filled.append(NULL);
}


// Used by BlockCollector::inputLoop to add a Subband to a Block into a BlockCollector.
// If this completes a Block (all Subbands for this Block received) and no
// subsequent Blocks are missing something, send it (or them) off into the
// outputQueue for write-back to storage.
void BlockCollector::processSubband( SmartPtr<Subband> &subband ) {
  ++nrBlockSubbandsReceived;

  LOG_DEBUG_STR(logPrefix << "Add " << subband->id);

  const size_t &blockIdx = subband->id.block;

  ASSERT(nrBlocks == 0 || blockIdx < nrBlocks);

  if (!have(blockIdx)) {
    if (!fetch(blockIdx)) {
      // too late -- discard packet
      ++nrBlockSubbandsTooLate;
      LOG_DEBUG_STR(logPrefix << "Dropped subband " << subband->id.subband);

      // if we can't drop, we shouldn't even be here.
      ASSERTSTR(canDrop, "Received block " << blockIdx << ", but already emitted up to " << lastEmitted << " for stream " << subband->id.fileIdx << " subband " << subband->id.subband);

      return;
    }
  }

  SmartPtr<Block> &block = blocks.at(blockIdx);

  block->addSubband(subband);

  if (block->complete()) {
    // Block is complete -- send it downstream,
    // and everything before it. We know we won't receive
    // data from earlier blocks, because each subband
    // is sent in-order.
    //
    // Note that blocks can be received out-of-order once
    // (many) subbands go missing. 
    emitUpTo(blockIdx);

    if (nrBlocks > 0 && blockIdx == nrBlocks - 1) {
      // Received last block -- wrap up
      LOG_INFO_STR(logPrefix << "Received last block");

      ASSERT(blocks.empty());

      // Signal end-of-stream
      outputQueue.noMore();
    }
  }
}


void BlockCollector::finish() {
  // Wait for all input to be processed
  inputQueue.noMore();
  inputThread.wait();

  // Wrap-up remainder
  if (!blocks.empty()) {
    emitUpTo(maxBlock());
  }

  // Update loss statistics
  nrFullBlocksLost += nrBlocks - lastEmitted + 1;
  if (nrBlocks > 0) {
    const float didNotReceivePerc = 100.0 - 100.0*nrBlockSubbandsReceived/nrSubbands/nrBlocks;
    const float didNotSendPerc = (100.0*nrFullBlocksLost + 100.0*nrBlockSubbandsLost/nrSubbands) / nrBlocks;
    const float lostPerc = didNotSendPerc - didNotReceivePerc;
    const float tooLatePerc = 100.0 * nrBlockSubbandsTooLate / nrSubbands / nrBlocks;

    if (didNotReceivePerc > 0)
      LOG_WARN_STR(logPrefix << "Did not receive " << didNotReceivePerc << "% of the data");
    if (lostPerc > 0)
      LOG_ERROR_STR(logPrefix << "I lost " << lostPerc << "% of the data");
    if (didNotSendPerc > 0)
      LOG_WARN_STR(logPrefix << "Did not send " << didNotSendPerc << "% of the data");
    if (tooLatePerc > 0)
      LOG_ERROR_STR(logPrefix << "Received " << tooLatePerc << "% of the data too late. Consider increasing maxBlocksInFlight.");
  }

  if (!canDrop) {
    // Should have received everything
    ASSERT(nrBlocks == 0 || (ssize_t)nrBlocks == lastEmitted + 1);
  }

  // Signal end-of-stream
  outputQueue.noMore();
  outputThread.wait();
}


size_t BlockCollector::minBlock() const {
  ASSERT(!blocks.empty());

  return blocks.begin()->first;
}

size_t BlockCollector::maxBlock() const {
  ASSERT(!blocks.empty());

  return blocks.rbegin()->first;
}


// Send the Block with blockIdx for write-back to storage now.
void BlockCollector::emit(size_t blockIdx) {
  // should emit in-order
  if (!canDrop) {
    ASSERT((ssize_t)blockIdx == lastEmitted + 1);
  } else {
    ASSERT((ssize_t)blockIdx > lastEmitted);
  }

  // fetch data
  SmartPtr<Block> &block = blocks.at(blockIdx);
  LOG_DEBUG_STR(logPrefix << "Emitting block " << blockIdx << " (last emit: " << lastEmitted << ")");

  // update loss statistics
  nrFullBlocksLost += blockIdx - lastEmitted + 1;
  nrBlockSubbandsLost += block->nrSubbandsLeft;
  
  // emit to outputPool.filled()
  outputQueue.append(block);

  // remove from our administration
  blocks.erase(blockIdx);

  // remember where we left off
  lastEmitted = blockIdx;
}


void BlockCollector::emitUpTo(size_t block) {
  while (!blocks.empty() && minBlock() <= block) {
    emit(minBlock());
  }
}


bool BlockCollector::have(size_t block) const {
  return blocks.find(block) != blocks.end();
}


bool BlockCollector::fetch(size_t block) {
  if ((ssize_t)block <= lastEmitted)
    // too late -- discard packet
    return false;

  // Make sure we don't exceed our maximum cache size
  if (canDrop && blocks.size() >= maxBlocksInFlight) {
    // Under severe data loss, we can get disjunct sets
    // of subbands for each block. In that case,
    // the blocks can arrive out-of-order.
    //
    // We should not accept any blocks that are even earlier
    // than the one we're about to emit.
    if (block < minBlock())
      return false;

    // No more room -- force out oldest block
    emit(minBlock());
  }

  // Add and annotate
  ASSERT(!have(block));
  blocks[block] = new Block(fileIdx, block, nrSubbands, nrSamples, nrChannels);

  return true;
}


// Receives data from an rtcp process. Runs in outputProc.
Receiver::Receiver( Stream &stream, CollectorMap &collectors )
:
  stream(stream),
  collectors(collectors),
  thread(this, &Receiver::receiveLoop)
{
}


Receiver::~Receiver()
{
  kill();
}


void Receiver::kill()
{
  thread.cancel();
}


bool Receiver::finish()
{
  thread.wait();

  for (CollectorMap::iterator i = collectors.begin(); i != collectors.end(); ++i)
    i->second->finish();

  return !thread.caughtException();
}


void Receiver::receiveLoop()
{
  try {
    for(;;) {
      SmartPtr<Subband> subband = new Subband;

      subband->read(stream);

      const size_t fileIdx = subband->id.fileIdx;

      ASSERTSTR(collectors.find(fileIdx) != collectors.end(), "Received a piece of stream " << fileIdx << ", which is unknown to me");

      //LOG_DEBUG_STR("File " << fileIdx << ": Adding subband " << subband.id);

      collectors.at(fileIdx)->addSubband(subband);
    }
  } catch (EndOfStreamException &) {
  }
}


// Listen for and receive connections and data in an outputProc process
// from MultiSender objects at different rtcp processes.
MultiReceiver::MultiReceiver( const std::string &servicePrefix, Receiver::CollectorMap &collectors )
:
  servicePrefix(servicePrefix),
  collectors(collectors),
  thread(this, &MultiReceiver::listenLoop)
{
}


// cancel listeners
MultiReceiver::~MultiReceiver()
{
  kill(0);
}


void MultiReceiver::kill(size_t minNrClients)
{
  {
    ScopedLock sl(mutex);

    while(clients.size() < minNrClients) {
      LOG_DEBUG_STR("MultiReceiver::kill: " << clients.size() << " clients connected, waiting for " << minNrClients);
      newClient.wait(mutex);
    }

  }

  // Kill listenLoop, but keep clients alive
  thread.cancel();

  // Wait for thread to die so we can access `clients'
  // safely.
  thread.wait();

  // Kill all client connections
  for (size_t i = 0; i < clients.size(); ++i) {
    if (minNrClients == 0) {
      clients[i].receiver->kill();
    } else {
      clients[i].receiver->finish();
    }
  }
}


// Listen for incoming connects on outputProc from rtcp.
void MultiReceiver::listenLoop()
{
  for(;;) {
    // Accept a new client
    SmartPtr<PortBroker::ServerStream> stream;
   
    try {
      stream = new PortBroker::ServerStream(servicePrefix, true);
    } catch(TimeOutException &) {
      // fail silently if no client connected
      LOG_DEBUG_STR("TABTranspose::MultiReceiver: Timed out");
      break;
    }

    LOG_DEBUG_STR("TABTranspose::MultiReceiver: Accepted resource " << stream->getResource());

    // Dispatch a Receiver for the client
    dispatch(stream.release());
  }
}


// Hook up a new Receiver (connected to rtcp) to block collectors,
// both in outputProc.
void MultiReceiver::dispatch( PortBroker::ServerStream *stream )
{
  ScopedLock sl(mutex);

  struct Client client;

  client.stream = stream;
  client.receiver = new Receiver(*stream, collectors);

  clients.push_back(client);

  // Inform of change in collectors
  newClient.signal();

  LOG_DEBUG_STR("TABTranspose::MultiReceiver: Dispatched client for resource " << stream->getResource());
}


// Maintains the connections of an rtcp process with all its outputProc processes
// it needs to send data to.
MultiSender::MultiSender( const HostMap &hostMap, const Parset &parset,
                          double maxRetentionTime, const std::string &bind_local_iface )
:
  hostMap(hostMap),
  itsParset(parset),
  maxRetentionTime(maxRetentionTime),
  bind_local_iface(bind_local_iface)
{
  for (HostMap::const_iterator i = hostMap.begin(); i != hostMap.end(); ++i) {
    // keep a list of unique hosts
    if(find(hosts.begin(), hosts.end(), i->second) == hosts.end())
      hosts.push_back(i->second);

    // each host gets a drop_rate counter
    drop_rates[i->second] = RunningStatistics("%");
  }

  for (vector<struct Host>::const_iterator i = hosts.begin(); i != hosts.end(); ++i) {
    queues[*i] = new Queue< SmartPtr<struct Subband> >(str(format("MultiSender::queue [to %s]") % i->hostName));
  }
}


MultiSender::~MultiSender()
{
  // at this point, process() has finished, either gracefully or with an exception

  // drain queues in case process() did not
  for (std::map<Host, QueuePtr>::iterator i = queues.begin(); i != queues.end(); ++i) {
    QueuePtr &queue = i->second;

    while (!queue->empty()) {
      (void)queue->remove();

      // this is loss, obviously
      drop_rates.at(i->first).push(100.0);
    }
  }

  // report statistics
  for (std::map<Host, RunningStatistics>::const_iterator i = drop_rates.begin(); i != drop_rates.end(); ++i) {
    const float didNotReceivePerc = 0.0; // no output loss possible so far
    const float lostPerc = i->second.mean();
    const float didNotSendPerc = didNotReceivePerc + lostPerc;

    if (lostPerc > 0)
      LOG_ERROR_STR("[BeamFormer] I lost " << lostPerc << "% of the data to " << i->first.hostName);
    if (didNotSendPerc > 0)
      LOG_WARN_STR("[BeamFormer] Did not send " << lostPerc << "% of the data to " << i->first.hostName);
  }

}


// Sets up the connections from an rtcp process to its outputProc processes.
// Then, keep writing blocks until we see a NULL Block ptr.
void MultiSender::process( OMPThreadSet *threadSet )
{
  // We need to register our threads somewhere...
  OMPThreadSet dummySet;

  if (!threadSet)
    threadSet = &dummySet;

#pragma omp parallel for num_threads(hosts.size())
  for (int i = 0; i < (ssize_t)hosts.size(); ++i) {
    const struct Host &host = hosts[i];
    const string logPrefix = str(format("[MultiSender -> %s] ") % host.hostName);

    try {
      OMPThreadSet::ScopedRun sr(*threadSet);

      LOG_DEBUG_STR(logPrefix << "Connecting to " << host.hostName << ":" << host.brokerPort << ":" << host.service);

      PortBroker::ClientStream stream(host.hostName, host.brokerPort, host.service, 0, bind_local_iface);

      LOG_DEBUG_STR(logPrefix << "Connected");

      QueuePtr &queue = queues.at(host);

      SmartPtr<struct Subband> subband;
      NSTimer sendTimer(str(format("Send Subband to %s") % host.hostName), true, true);

      while ((subband = queue->remove()) != NULL) {
        NSTimer::StartStop ss(sendTimer);

        // Note: this can take any amount of time, even hours, even in real-time mode
        subband->write(stream);
      }

      LOG_DEBUG_STR(logPrefix << "Done");
    } catch (SystemCallException &ex) {
      LOG_ERROR_STR(logPrefix << "Caught exception: " << ex.what());
    } catch (Exception &ex) {
      LOG_ERROR_STR(logPrefix << "Caught exception: " << ex);
    }
  }
}


// The pipeline calls here to write a block for a single file (part).
bool MultiSender::append( SmartPtr<struct Subband> &subband )
{
  using namespace TimeSpec;

  // Find the host to send these data to
  const size_t fileIdx = subband->id.fileIdx;
  const struct Host &host = hostMap.at(fileIdx);

  QueuePtr &queue = queues.at(host);

  bool dropped = false;

  // If oldest packet in queue is too old, drop it in lieu of this new one
  if (itsParset.settings.realTime && TimeSpec::now() - queue->oldest() > maxRetentionTime) {
    drop_rates.at(host).push(100.0);

    // remove oldest item
    SmartPtr<struct Subband> subband = queue->remove();

    // would be weird to have NULL in here while we're appending elements
    ASSERT(subband);

    dropped = true;
  } else {
    drop_rates.at(host).push(0.0);
  }

  // Append the data to the respective queue
  queue->append(subband);

  return !dropped;
}


// The pipeline indicates that it has appended its last block.
// Further appending is denied and a final NULL Block ptr is queued.
void MultiSender::finish()
{
  for (vector<struct Host>::const_iterator i = hosts.begin(); i != hosts.end(); ++i) {
    queues.at(*i)->append(NULL, false);
  }
}

} // namespace TABTranspose
} // namespace Cobalt
} // namespace LOFAR

