//# MSWriterCorrelated.cc: a writer for correlated visibilities
//# Copyright (C) 2012-2013  ASTRON (Netherlands Institute for Radio Astronomy)
//# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
//#
//# This file is part of the LOFAR software suite.
//# The LOFAR software suite is free software: you can redistribute it and/or
//# modify it under the terms of the GNU General Public License as published
//# by the Free Software Foundation, either version 3 of the License, or
//# (at your option) any later version.
//#
//# The LOFAR software suite is distributed in the hope that it will be useful,
//# but WITHOUT ANY WARRANTY; without even the implied warranty of
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//# GNU General Public License for more details.
//#
//# You should have received a copy of the GNU General Public License along
//# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
//#
//# $Id$

#include <lofar_config.h>

#include "MSWriterCorrelated.h"

#include <sys/types.h>
#include <fcntl.h>
#include <cstdio>
#include <cstdlib>
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>

#include <Common/SystemUtil.h>
#include <Common/Timer.h>
#include <Common/Thread/Mutex.h>
#include <MSLofar/FailedTileInfo.h>
#include <CoInterface/CorrelatedData.h>
#include <CoInterface/LTAFeedback.h>

#include <tables/Tables/Table.h>
#include <casa/Quanta/MVTime.h>

#include "MeasurementSetFormat.h"

using boost::format;
using namespace casa;


namespace LOFAR
{
  namespace Cobalt
  {
    // protect system("mv") with a mutex to allow us to time it across threads
    Mutex mvMutex;
    NSTimer mvTimer("MeasurementSet moving to target file system", true, true);
    NSTimer finaliseTimer("MeasurementSet finalisation (adding broken antenna info)", true, true);

    static std::string stagePath(const std::string &origPath, const std::string &stageDir)
    {
      // no staging dir means use original location
      if (stageDir.empty())
        return origPath;

      // use filename from origPath, directory from stageDir
      return str(format("%s/%s") % stageDir % basename(origPath));
    }

    MSWriterCorrelated::MSWriterCorrelated (const std::string &logPrefix, const std::string &msName, const Parset &parset, unsigned subbandIndex)
      :
      // Write the data to the staging dir, until we have a MeasurementSet we can move it into.
      MSWriterFile(stagePath(str(format("%s-table.f0data") % msName), parset.settings.correlator.files[subbandIndex].dataStagingDirTargetFS)),
      itsLogPrefix(logPrefix),
      itsMSname(msName),
      itsParset(parset),
      itsSubbandIndex(subbandIndex)
    {
      // Add file-specific processing feedback
      LTAFeedback fb(itsParset.settings);
      itsConfiguration.adoptCollection(fb.correlatedFeedback(itsSubbandIndex));
      itsConfigurationPrefix = fb.correlatedPrefix(itsSubbandIndex);

      // Create Sequence file (in the staging dir)
      if (LofarStManVersion > 1) {
        string seqfilename = stagePath(str(format("%s-table.f0seqnr") % msName), parset.settings.correlator.files[subbandIndex].dataStagingDirTargetFS);

        try {
          itsSequenceNumbersFile = new FileStream(seqfilename, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
        } catch (...) {
          LOG_WARN_STR(itsLogPrefix << "Could not open sequence numbers file " << seqfilename);
        }
      }
    }


    void MSWriterCorrelated::init()
    {
      // Create MeasurementSet in "metaDataStagingDirCreateFS", where it can be created efficiently
#if defined HAVE_AIPSPP
      MeasurementSetFormat myFormat(itsParset, 512);

      const std::string createPathMS = stagePath(itsMSname, itsParset.settings.correlator.files[itsSubbandIndex].metaDataStagingDirCreateFS);

      LOG_INFO_STR(itsLogPrefix << "Creating MeasurementSet " << createPathMS);
      myFormat.addSubband(createPathMS, itsSubbandIndex);

      // Move the MeasurementSet to "metaDataStagingDirTargetFS", to stage it on the target file system
      const std::string movePathMS = stagePath(itsMSname, itsParset.settings.correlator.files[itsSubbandIndex].metaDataStagingDirTargetFS);
      if (createPathMS != movePathMS) {
        ScopedLock sl(mvMutex);
        NSTimer::StartStop ss(mvTimer);

        LOG_INFO_STR(itsLogPrefix << "Moving MeasurementSet " << createPathMS << " to " << movePathMS);

        const std::string mv_cmd = str(format("mv -fT '%s' '%s'") % createPathMS % movePathMS);
        if (::system(mv_cmd.c_str()) != 0)
          THROW(StorageException, str(format("system(%s)") % mv_cmd));
      }

      // Move the MeasurementSet to its final location
      recursiveMakeDir(dirname(itsMSname), itsLogPrefix);

      if (rename(movePathMS.c_str(), itsMSname.c_str()) < 0)
        THROW_SYSCALL(str(format("rename(%s, %s)") % movePathMS % itsMSname));

      LOG_DEBUG_STR(itsLogPrefix << "MeasurementSet created");
#endif // defined HAVE_AIPSPP

      // Move data file into the measurement set
      if (rename(itsFile.name.c_str(),
                 str(format("%s/table.f0data") % itsMSname).c_str()) < 0)
        THROW_SYSCALL(str(format("rename(%s, %s/table.f0data)") % itsFile.name % itsMSname));

      // Move sequence file into the measurement set
      if (rename(itsSequenceNumbersFile->name.c_str(),
                 str(format("%s/table.f0seqnr") % itsMSname).c_str()) < 0)
        THROW_SYSCALL(str(format("rename(%s, %s/table.f0seqnr)") % itsSequenceNumbersFile->name % itsMSname));
    }


    MSWriterCorrelated::~MSWriterCorrelated()
    {
    }


    void MSWriterCorrelated::write(StreamableData *data)
    {
      CorrelatedData *cdata = dynamic_cast<CorrelatedData*>(data);

      ASSERT( data );
      ASSERT( cdata );

      // Write data
      MSWriterFile::write(data);

      // Write sequence number
      if (itsSequenceNumbersFile != 0) {
        // quick fix: always write to maintain integrity
        unsigned seqnr = data->sequenceNumber(true);

        itsSequenceNumbersFile->write(&seqnr, sizeof seqnr);
      }

      itsNrBlocksWritten++;

      itsConfiguration.replace(itsConfigurationPrefix + "size",     str(format("%u") % getDataSize()));
      itsConfiguration.replace(itsConfigurationPrefix + "duration", str(format("%f") % ((data->sequenceNumber() + 1) * itsParset.settings.correlator.integrationTime())));
      itsConfiguration.replace(itsConfigurationPrefix + "percentageWritten", str(format("%u") % percentageWritten()));
    }


    static MVEpoch datetime2epoch(const string &datetime)
    {
      Quantity q;

      if (!MVTime::read(q, datetime))
        return MVEpoch(0);

      return MVEpoch(q);
    }


    void MSWriterCorrelated::fini(const FinalMetaData &finalMetaData)
    {
      ScopedLock sl(MeasurementSetFormat::sharedMutex);
      NSTimer::StartStop ss(finaliseTimer);

      map<string, FailedTileInfo::VectorFailed> brokenBefore, brokenDuring;

      // fill set of broken hardware at beginning of observation
      for (size_t i = 0; i < finalMetaData.brokenRCUsAtBegin.size(); i++) {
        const struct FinalMetaData::BrokenRCU &rcu = finalMetaData.brokenRCUsAtBegin[i];

        brokenBefore[rcu.station].push_back(FailedTileInfo(rcu.station, rcu.time, datetime2epoch(rcu.time), rcu.type, rcu.seqnr));
      }

      // fill set of hardware that broke during the observation
      for (size_t i = 0; i < finalMetaData.brokenRCUsDuring.size(); i++) {
        const struct FinalMetaData::BrokenRCU &rcu = finalMetaData.brokenRCUsDuring[i];

        brokenDuring[rcu.station].push_back(FailedTileInfo(rcu.station, rcu.time, datetime2epoch(rcu.time), rcu.type, rcu.seqnr));
      }

      LOG_DEBUG_STR(itsLogPrefix << "Reopening MeasurementSet");

      Table ms(itsMSname, Table::Update);

      vector<FailedTileInfo::VectorFailed> before(FailedTileInfo::antennaConvert(ms, brokenBefore));
      vector<FailedTileInfo::VectorFailed> during(FailedTileInfo::antennaConvert(ms, brokenDuring));

      LOG_DEBUG_STR(itsLogPrefix << "Writing broken hardware information to MeasurementSet");

      try {
        FailedTileInfo::writeFailed(ms, before, during);
      } catch (Exception &ex) {
        LOG_ERROR_STR("Failed to write broken hardware information: " << ex);
      }
    }


  } // namespace Cobalt
} // namespace LOFAR

