//# TBB_Dipole.cc: TBB per-dipole routines to store incoming TBB data
//# Copyright (C) 2012-2017  ASTRON (Netherlands Institute for Radio Astronomy)
//# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
//#
//# This file is part of the LOFAR software suite.
//# The LOFAR software suite is free software: you can redistribute it and/or
//# modify it under the terms of the GNU General Public License as published
//# by the Free Software Foundation, either version 3 of the License, or
//# (at your option) any later version.
//#
//# The LOFAR software suite is distributed in the hope that it will be useful,
//# but WITHOUT ANY WARRANTY; without even the implied warranty of
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//# GNU General Public License for more details.
//#
//# You should have received a copy of the GNU General Public License along
//# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
//#
//# $Id$

#include <lofar_config.h>

#include "TBB_Dipole.h"
#include <cstring>
#include <unistd.h>
#include <endian.h>
#if __BYTE_ORDER != __BIG_ENDIAN && __BYTE_ORDER != __LITTLE_ENDIAN
#error Byte order is neither big endian nor little endian: not supported
#endif
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>

#ifdef basename // some glibc have this as a macro
#undef basename
#endif
#include <Common/SystemUtil.h>
#include <Common/LofarLogger.h>
#include <Common/SystemCallException.h>

namespace LOFAR
{
  namespace Cobalt
  {

    using namespace std;

    // LCS/Stream FileStream does not wrap pwrite(2). No other LOFAR source uses it, so define it here in the same way.
    static size_t tryPWrite(int fd, const void *ptr, size_t size, off_t offset)
    {
      ssize_t bytes = ::pwrite(fd, ptr, size, offset);
      if (bytes < 0)
        THROW_SYSCALL("pwrite");
      return bytes;
    }

    static void pwrite(int fd, const void *ptr, size_t size, off_t offset)
    {
      while (true) {
        size_t bytes = tryPWrite(fd, ptr, size, offset);
        size -= bytes;
        if (size == 0) {
          return;
        }
        offset += bytes;
        ptr = static_cast<const char *>(ptr) + bytes;
      }
    }


    TBB_Dipole::TBB_Dipole()
      : itsDataset(),
        itsLastLogErrorTime(0)
    {
    }

    // Do not use. Only needed for vector<TBB_Dipole>(N).
    TBB_Dipole::TBB_Dipole(const TBB_Dipole& rhs)
      : itsDataset(),
        itsLastLogErrorTime(rhs.itsLastLogErrorTime)
    {
    }

    TBB_Dipole::~TBB_Dipole()
    {
      // Executed by the main thread after joined with all workers, so no need to lock or delay cancellation.
      if (isInitialized()) {
        unsigned sb;
        if (doTransient()) {
          sb = 0;
        } else { // spectral mode
          // TODO: As long as we don't know dump duration / length beforehand, the best we can do here is write from the 1st used itsDumpInfo element. 
          for (sb = 0; sb < itsDumpInfo.size(); sb++) {
            if (!!itsDumpInfo[sb].itsRawFile) { // initialized
              break; // found
            }
          }
          if (sb == RSP_NR_SUBBANDS) { // not found: ensure attribs below can be written
            sb = 0;
            itsDumpInfo[sb].itsDatasetLen = 0;
          }
        }

        try {
          if (doTransient()) {
            itsDataset->resize1D(itsDumpInfo[sb].itsDatasetLen);
          } else { // spectral mode
            itsDataset->resize1D(itsDumpInfo[sb].itsDatasetLen * 2); // TODO: remove times 2 once DAL supports TBB spectral dataset of complex<int16_t>. Also, use resize() if switched to create() (2D).
          }
        } catch (exception& exc) { // dal::DALException, or std::bad_alloc from vector constr
          LOG_WARN("TBB: failed to resize HDF5 dipole dataset to external data size: " + string(exc.what()));
        }

        try {
          // Thus in values (scalar for transient mode, complex for subband mode)
          itsDataset->dataLength().value = itsDumpInfo[sb].itsDatasetLen;
        } catch (dal::DALException& exc) {
          LOG_WARN("TBB: failed to set dipole DATA_LENGTH attribute: " + string(exc.what()));
        }
        try {
          /*
           * We write flags at the end, because HDF5 attributes can only be (re)set, not updated.
           * TODO: If an .h5 internal (non-chunked!) 1D dataset would be possible without pre-specifying size,
           * that would be better. Then we can write flags while data comes in. Else, leave as-is.
           */
          itsDataset->flagOffsets().create(itsDumpInfo[sb].itsFlagOffsets.size()).set(itsDumpInfo[sb].itsFlagOffsets);
        } catch (dal::DALException& exc) {
          LOG_WARN("TBB: failed to set dipole FLAG_OFFSETS attribute: " + string(exc.what()));
        }
      }
    }

    void TBB_Dipole::init(const TBB_Header& header, const Parset& parset,
                          const StationMetaData& stationMetaData,
                          const vector<double>& allSubbandCentralFreqs, const string& h5Filename,
                          dal::TBB_Station& station, Mutex& h5Mutex)
    {
      const bool doTransientMode = allSubbandCentralFreqs.empty(); // cannot use doTransient() yet
      const string rawFilename = getRawFilename(h5Filename, doTransientMode,
                                                header.rspID, header.rcuID);
      string filename;

      if (doTransientMode) {
        itsDumpInfo.resize(1);
        filename = rawFilename;
        itsDumpInfo[0].itsRawFile = SmartPtr<FileStream>(new FileStream(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH));
        itsDumpInfo[0].itsDatasetLen = 0;
        itsDumpInfo[0].itsTime0 = header.time;
        itsDumpInfo[0].itsSampleNrOrSliceNr0 = getTransientModeSampleNr(header);
      } else { // spectral mode 
        /*
         * For spectral mode, we store 1 subband per file (until LOFAR control writes dump duration or length in the parset).
         * DAL is not ready yet for spectral mode and takes 1 raw filename.
         * For now, iterate backwards to pass the 1st raw filename to createTBB_DipoleDataset().
         */
        itsDumpInfo.resize(RSP_NR_SUBBANDS); // preallocate all is easy, safe and good enough
        for (unsigned sb = allSubbandCentralFreqs.size(); sb-- > 0; ) {
          if (allSubbandCentralFreqs[sb] == 0.0) {
            continue; // subband not dumped (according to parset)
          }
          filename = str(boost::format(rawFilename) % sb);
          itsDumpInfo[sb].itsRawFile = SmartPtr<FileStream>(new FileStream(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH));
          itsDumpInfo[sb].itsDatasetLen = 0;
          // itsDumpInfo[sb].itsTime0 is lazily initialized later (cannot init for all subbands with 1st packet)
          // itsDumpInfo[sb].itsSampleNrOrSliceNr0 is laziliy initialized later (idem)
        }
      }

      {
        ScopedLock h5OutLock(h5Mutex);
        initTBB_DipoleDataset(header, parset, stationMetaData,
                              allSubbandCentralFreqs, filename, station);
      }
    }

    bool TBB_Dipole::isInitialized() const
    {
      return !!itsDataset;
    }

    bool TBB_Dipole::doTransient() const
    {
      return itsDumpInfo.size() == 1;
    }

    string TBB_Dipole::getRawFilename(const string& h5Filename, bool doTransient,
                                      unsigned rspID, unsigned rcuID)
    {
      string rawFilename = h5Filename;
      string rspRcuSbStr = str(boost::format("_RSP%03u_RCU%03u") % rspID % rcuID);
      if (!doTransient) {
        rspRcuSbStr.append("_SB%03u"); // insert conv specifier for subband nr in spectral mode (can be removed when all stored in 1 file)
      }
      size_t pos = rawFilename.find('_', rawFilename.find('_') + 1);
      rawFilename.insert(pos, rspRcuSbStr); // insert _RSPxxx_RCUxxx after station name (2nd '_')
      rawFilename.resize(rawFilename.size() - (sizeof(".h5") - 1));
      rawFilename.append(".raw");
      return rawFilename;
    }

    // Add a new flag range at the end or extend the last stored flag range. 'len' may not be 0.
    void TBB_Dipole::appendFlags(DumpInfo& di, uint64_t offset, size_t len)
    {
      if (di.itsFlagOffsets.empty() || offset > di.itsFlagOffsets.back().end) {
        di.itsFlagOffsets.push_back(dal::Range(offset, offset + len));
      } else { // extend
        di.itsFlagOffsets.back().end += len;
      }
    }

    uint32_t TBB_Dipole::getTransientModeSampleNr(const TBB_Header& header) const
    {
      /*
       * In transient mode, at 200 MHz we get DEFAULT_TBB_TRANSIENT_NSAMPLES (1024) samples per frame: 195312.5 frames/s.
       * This means that every 2 seconds, a frame overlaps a seconds boundary. But the sample values generated
       * by the RSPs start at zero for each second, even if it should start at 512 for odd timestamps at 200 MHz.
       * At 160 MHz sample rate, an integer number of frames fits in a second (156250), so no correction is needed.
       */
      uint32_t sampleNr = header.sampleNrOrBandSliceNr;
      if (header.sampleFreq == 200 && header.time & 1) {
        sampleNr += DEFAULT_TBB_TRANSIENT_NSAMPLES / 2;
      }
      return sampleNr;
    }

    void TBB_Dipole::processTransientFrameData(const TBB_Frame& frame)
    {
      /*
       * Out-of-order frame arrival has not been seen for Dutch stations. TBB from int'l stations is not (yet) dumped to CEP.
       * We could support out-of-order arrival, except for packets that appear to be overtaken by the 1st packet received,
       * but this makes flagging lost packets more complicated than how we deal with it using appendFlags() below.
       * Also, fake out-of-order occurs when data is dumped (incorrectly) across the frozen TBB write pointer (LOFAR control bug).
       */
      DumpInfo& di = itsDumpInfo[0];
      uint32_t sampleNr = getTransientModeSampleNr(frame.header);
      if (frame.header.time < di.itsTime0 ||
          (frame.header.time == di.itsTime0 && sampleNr < di.itsSampleNrOrSliceNr0)) {
        logErrorRateLimited(&itsLastLogErrorTime, "TBB: unhandled out-of-order packet: " +
                                                  frame.header.to_string());
        return;
      }

      uint64_t offset = (frame.header.time - di.itsTime0) * frame.header.sampleFreq * 1000000;
      offset += sampleNr - di.itsSampleNrOrSliceNr0;

      /*
       * Flag lost frame(s) (assume no out-of-order).
       * Assumes all frames (except maybe the last) have the same nr of samples.
       * This cannot detect lost frames at the end of a dataset.
       */
      int64_t nskipped = offset - di.itsDatasetLen; // should be > 0, but do signed cmp to avoid crazy flagging range
      if (nskipped > 0) {
        appendFlags(di, di.itsDatasetLen, (uint64_t)nskipped);
      }

      // On a data checksum error or all zeros, flag these samples.
      if (!crc32tbb(&frame.payload, frame.header.nOfSamplesPerFrame)) {
        appendFlags(di, offset, frame.header.nOfSamplesPerFrame);
        uint32_t crc32;
        memcpy(&crc32, &frame.payload.data[frame.header.nOfSamplesPerFrame], sizeof crc32); // strict-aliasing safe
        logErrorRateLimited(&itsLastLogErrorTime, "TBB: crc32 error: " + frame.header.to_string() +
                                                  " crc32: " + boost::lexical_cast<string>(crc32));
      } else if (hasAllZeroDataSamples(frame.payload, frame.header.nOfSamplesPerFrame)) {
        appendFlags(di, offset, frame.header.nOfSamplesPerFrame);
      }

      // Since we are writing around HDF5, there is no need to lock. Resize the HDF5 dataset in the destructor.
      pwrite(di.itsRawFile->fd, frame.payload.data, frame.header.nOfSamplesPerFrame * sizeof(int16_t),
             (off_t)(offset * sizeof(int16_t)));
      di.itsDatasetLen = offset + frame.header.nOfSamplesPerFrame;
    }

    void TBB_Dipole::processSpectralFrameData(const TBB_Frame& frame)
    {
      // NOTE: only intends to support TBB-ALERT subband mode, and normal subband mode with 1 band per packet!

      /*
       * NOTE: frame header bandSel may have multiple bits set.
       * Atm, we don't support this, because the (other) bandNr sub-field in the header appears odd wrt the spec
       * (e.g. 0 even if band 0 is not dumped (maybe it's an index?)) for multi-subband dumps.
       * The TBB-ALERT subband mode firmware does not send (parts of) multiple subbands per packet anyway.
       */
      unsigned bandSelNr = frame.header.getFirstBandSelNr();
      if (bandSelNr == RSP_NR_SUBBANDS) {
        logErrorRateLimited(&itsLastLogErrorTime, "TBB: dropping packet: subband mode, but empty bandSel bitmap in packet header");
        return;
      }
      DumpInfo& di = itsDumpInfo[bandSelNr];
      if (!di.itsRawFile) { // not initialized: band not to be expected according to parset
        logErrorRateLimited(&itsLastLogErrorTime, "TBB: dropping packet: packet subband nr not in parset Observation.TBB.TBBsetting.subbandList: " +
                                                  boost::lexical_cast<string>(bandSelNr));
        return;
      }

      // Lazily initialize last part of dump info struct from 1st frame received
      if (di.itsTime0 == 0) {
        di.itsTime0 = frame.header.time;
        di.itsSampleNrOrSliceNr0 = frame.header.sampleNrOrBandSliceNr >> TBB_SLICE_NR_SHIFT;
      }

      /*
       * Out-of-order frame arrival has not been seen for Dutch stations. TBB from int'l stations is not (yet) dumped to CEP.
       * We could support out-of-order arrival, except for packets that appear to be overtaken by the 1st packet received,
       * but this makes flagging lost packets more complicated than how we deal with it using appendFlags() below.
       * Also, fake out-of-order occurs when data is dumped (incorrectly) across the frozen TBB write pointer (LOFAR control bug).
       */
      uint32_t sliceNr = frame.header.sampleNrOrBandSliceNr >> TBB_SLICE_NR_SHIFT;
      if (frame.header.time < di.itsTime0 ||
          (frame.header.time == di.itsTime0 && sliceNr < di.itsSampleNrOrSliceNr0)) {
        logErrorRateLimited(&itsLastLogErrorTime, "TBB: unhandled out-of-order packet: " +
                                                  frame.header.to_string());
        return;
      }

      // With a 200 MHz sample freq, even seconds have 1 extra slice (that actually falls half-way).
      uint64_t offset = (frame.header.time - di.itsTime0) * frame.header.sampleFreq * 1000000 / SPECTRAL_TRANSFORM_SIZE;
      if (frame.header.sampleFreq == 200) { // NOTE: untested body: when can I have enough test data from a working firmware in subband mode?!?
        unsigned time0 = di.itsTime0;
        if (frame.header.time > time0 && !(time0 & 1)) {
          offset += 1; // 1st packet was in an even second and this packet is after that second: 1 extra slice fitted in that second
          time0 += 1; // avoid duplicate offset compensation next
        }
        offset += (frame.header.time - time0) / 2; // 1 extra slice every 2 seconds
      }
      offset += sliceNr - di.itsSampleNrOrSliceNr0;

      /*
       * Flag lost frame(s) (assumes no out-of-order).
       * Assumes all frames (except maybe the last) have the same nr of samples (fine).
       * This cannot detect lost frames at the end of a dataset.
       */
      int64_t nskipped = offset - di.itsDatasetLen; // should be > 0, but do signed cmp to avoid crazy flagging range
      if (nskipped > 0) {
        appendFlags(di, di.itsDatasetLen, (uint64_t)nskipped);
      }

      /*
       * On a data checksum error or all zeros, flag these samples.
       * TBB Design Doc states the crc32 is computed for transient data only,
       * but the firmware developer (Wietse Poiesz) says it is also valid for spectral data.
       * But in data dumps I've seen, it looked invalid for the first spectral frame each second; recheck this (?).
       */
      if (0) { // !crc32tbb(frame.payload.data, 2 * frame.header.nOfSamplesPerFrame)) {
        appendFlags(di, offset, frame.header.nOfSamplesPerFrame); // in (complex) values
        uint32_t crc32;
        memcpy(&crc32, &frame.payload.data[2 * frame.header.nOfSamplesPerFrame], sizeof crc32); // strict-aliasing safe
        logErrorRateLimited(&itsLastLogErrorTime, "TBB: crc32 error: " + frame.header.to_string() +
                                                  " crc32: " + boost::lexical_cast<string>(crc32));
      } else if (hasAllZeroDataSamples(frame.payload, 2 * frame.header.nOfSamplesPerFrame)) {
        appendFlags(di, offset, frame.header.nOfSamplesPerFrame); // in (complex) values
      }

      //unsigned bandNr = frame.header.sampleNrOrBandSliceNr & TBB_BAND_NR_MASK; // atm unused; from data it appears not to be a subband number; is it a subband idx?

      // NOTE: only intends to support TBB-ALERT subband mode, and normal subband mode with 1 band per packet!

      // Since we are writing around HDF5, there is no need to lock. Resize the HDF5 dataset in the destructor.
      pwrite(di.itsRawFile->fd, frame.payload.data, frame.header.nOfSamplesPerFrame * 2 * sizeof(int16_t),
             (off_t)(offset * 2 * sizeof(int16_t)));
      di.itsDatasetLen = offset + frame.header.nOfSamplesPerFrame; // in (complex) values
    }

    void TBB_Dipole::initTBB_DipoleDataset(const TBB_Header& header, const Parset& parset,
                                           const StationMetaData& stationMetaData,
                                           const vector<double>& /*allSubbandCentralFreqs*/,
                                           const string& rawFilename,
                                           dal::TBB_Station& station)
    {
      const string filename = LOFAR::basename(rawFilename); // don't store paths in HDF5

      // Override endianess. TBB data is always stored little endian and also received as such, so written as-is on any platform.
      if (doTransient()) {
        itsDataset.reset(new dal::TBB_DipoleDataset(station.dipole(header.stationID, header.rspID, header.rcuID)));
        itsDataset->create1D(0, -1, filename, itsDataset->LITTLE);

        itsDataset->sampleNumber().value = getTransientModeSampleNr(header);
      } else { // spectral mode
        itsDataset.reset(new dal::TBB_DipoleDataset(station.dipole(header.stationID, header.rspID, header.rcuID)));
/*        itsDataset = new dal::TBB_SubbandsDataset(station.subbands(header.stationID, header.rspID, header.rcuID)); TODO: something like this with new DAL version */
/* kept code around in case needed for when DAL properly supports subband mode
        vector<double> usedSubbandCentralFreqs;
        for (unsigned i = 0; i < allSubbandCentralFreqs.size(); i++) {
          if (allSubbandCentralFreqs[i] != 0.0) {
            usedSubbandCentralFreqs.push_back(allSubbandCentralFreqs[i]);
          }
        }

        vector<ssize_t> dims(2), maxdims(2);
        dims[0] = 0;
        dims[1] = usedSubbandCentralFreqs.size(); // create() will throw if this is 0
        maxdims[0] = -1; // only the 1st dim can be extendible
        maxdims[1] = usedSubbandCentralFreqs.size();
        itsDataset->create(dims, maxdims, filename, itsDataset->LITTLE);
*/
        itsDataset->create1D(0, -1, filename, itsDataset->LITTLE);
/* TODO: enable with new DAL version
        itsDataset->sliceNumber().value = header.sampleNrOrBandSliceNr >> TBB_SLICE_NR_SHIFT;
        itsDataset->spectralNofBands().value = usedSubbandCentralFreqs.size();
        itsDataset->spectralBands().create(usedSubbandCentralFreqs.size()).set(usedSubbandCentralFreqs);
        itsDataset->spectralBandsUnit().value = "MHz";
*/
      }
      LOG_INFO_STR("TBB: Created HDF5 Dataset for rsp,rcu " << (unsigned)header.rspID << ',' << (unsigned)header.rcuID << " for raw file " << filename);

      itsDataset->groupType().value = "DipoleDataset";
      itsDataset->stationID().value = header.stationID;
      itsDataset->rspID().value = header.rspID;
      itsDataset->rcuID().value = header.rcuID;

      itsDataset->sampleFrequency().value = header.sampleFreq;
      itsDataset->sampleFrequencyUnit().value = "MHz";

      itsDataset->time().value = header.time; // in seconds

      itsDataset->samplesPerFrame().value = header.nOfSamplesPerFrame; // possibly sanitized
      //itsDataset->dataLength().value is set at the end (destr)
      //itsDataset->flagOffsets().value is set at the end (destr)
      itsDataset->nyquistZone().value = parset.settings.nyquistZone();

      //#include "MAC/APL/PIC/RSP_Driver/src/CableSettings.h" or "RCUCables.h"
      // Cable delays (optional) from static meta data.
      //itsDataset->cableDelay().value = ???; // TODO
      //itsDataset->cableDelayUnit().value = "ns";

      /*
         > No DIPOLE_CALIBRATION_DELAY_VALUE
         > No DIPOLE_CALIBRATION_DELAY_UNIT
         These can be calculated from the values in the LOFAR calibration
         tables, but we can do that ourselves as long as the calibration table
         values for each dipole are written to the new keyword. Sander: please put them in; see the code ref below.
         DIPOLE_CALIBRATION_GAIN_CURVE.

         // Use StaticMetaData/CalTables

         calibration delay value en unit zijn nuttiger
         en is het beste om die er gelijk in te schrijven
         momenteel
         In /opt/cep/lus/daily/Mon/src/code/src/PyCRTools/modules/metadata.py
         heb ik code om de calibratie tabellen uit te lezen
         De functie: getStationPhaseCalibration
         elke .dat file bevat 96*512*2 doubles
         voor 96 rcus, 512 frequenties, een complexe waarde
         maar nu vraag ik me wel weer af of de frequenties of de rcus eerst komen
       */
      //NL stations: 768 kB, Int'l: 1.5 MB. Drop optional ASCI header. See also Station/StationCal/writeCalTable.m
      //itsDataset->dipoleCalibrationDelay().value = ???; // Pim can compute this from the GainCurve below
      //itsDataset->dipoleCalibrationDelayUnit().value = 's';
      //itsDataset->dipoleCalibrationGainCurve().create(???.size()).set(???); // st cal table
      //write cal tables into proper n-dimensional h5 data set, not attribute! Add access functions to DAL?

      // Skip if station is not participating in the observation (should not happen).
      if (stationMetaData.available && 2u * 3u * header.rcuID + 2u < stationMetaData.antPositions.size()) { // bounds check for antPositions[]
        /*TODO
         * Selecting the right positions depends on the antenna set. Checking vs the tables in
         * lhn001:/home/veen/lus/src/code/data/lofar/antennapositions/ can help, but their repos may be outdated.
         */
        vector<double> antPos(3);
        antPos[0] = stationMetaData.antPositions[2u * 3u * header.rcuID];
        antPos[1] = stationMetaData.antPositions[2u * 3u * header.rcuID + 1u];
        antPos[2] = stationMetaData.antPositions[2u * 3u * header.rcuID + 2u];
        itsDataset->antennaPosition().create(antPos.size()).set(antPos); // absolute position

        itsDataset->antennaPositionUnit().value = "m";
        itsDataset->antennaPositionFrame().value = parset.positionType(); // "ITRF"

        /*
         * The normal vector and rotation matrix are actually per antenna field,
         * but given the HBA0/HBA1 "ears" depending on antenna set, it was
         * decided to store them per antenna.
         */
        itsDataset->antennaNormalVector().create(stationMetaData.normalVector.size()).set(stationMetaData.normalVector);       // 3 doubles
        itsDataset->antennaRotationMatrix().create(stationMetaData.rotationMatrix.size()).set(stationMetaData.rotationMatrix); // 9 doubles, 3x3, row-major
      }

      // Tile beam is the analog beam. Only HBA can have one analog beam; optional.
      if (parset.settings.anaBeam.enabled) {
        vector<double> anaBeamDir(2);
        anaBeamDir[0] = parset.settings.anaBeam.direction.angle1;
        anaBeamDir[1] = parset.settings.anaBeam.direction.angle2;
        itsDataset->tileBeam().create(anaBeamDir.size()).set(anaBeamDir);
        itsDataset->tileBeamUnit().value = "m";
        itsDataset->tileBeamFrame().value = parset.settings.anaBeam.direction.type;

        //itsDataset->tileBeamDipoles().create(???.size()).set(???);

        //itsDataset->tileCoefUnit().value = ???;
        //itsDataset->tileBeamCoefs().value = ???;

        // Relative position within the tile.
        //itsDataset->tileDipolePosition().value = ???;
        //itsDataset->tileDipolePositionUnit().value = ???;
        //itsDataset->tileDipolePositionFrame().value = ???;
      }

      // TODO: TABs: support >1 TABS and add coh/incoh (this is for SAP 0, coh TAB 0), direction (incl type)
      double dpMeas;
      if (!parset.settings.beamFormer.anyCoherentTABs()) {
        dpMeas = 0.0;
      } else {
        dpMeas = -1.0;
        for (unsigned sap = 0; sap < parset.settings.beamFormer.SAPs.size(); sap++) {
          for (unsigned tab = 0; tab < parset.settings.beamFormer.SAPs[sap].TABs.size(); tab++) {
            const ObservationSettings::BeamFormer::TAB &t = parset.settings.beamFormer.SAPs[sap].TABs[tab];
            if (t.coherent) {
              dpMeas = t.dispersionMeasure;
              break;
            }
          }
          if (dpMeas != -1.0) {
            break;
          }
        }
      }
      itsDataset->dispersionMeasure().value = dpMeas;
      itsDataset->dispersionMeasureUnit().value = "pc/cm^3";
    }

    /*
     * NOTE: The nTrSamples arg is without the space taken by the crc32 in payload (drop too small frames earlier)
     * and in terms of the transient sample size, i.e. sizeof(int16_t).
     */
    bool TBB_Dipole::crc32tbb(const TBB_Payload* payload, size_t nTrSamples)
    {
      itsCrc32gen.reset();

      const char* ptr = reinterpret_cast<const char*>(payload->data); // to char* for strict-aliasing
      for (unsigned i = 0; i < nTrSamples * sizeof(int16_t); i += 2) {
        int16_t val;
        memcpy(&val, &ptr[i], sizeof val); // strict-aliasing safe
        val = __bswap_16(val);
        itsCrc32gen.process_bytes(&val, sizeof val);
      }

      // It is also possible to process crc32val and see if checksum() equals 0.
      uint32_t crc32val;
      memcpy(&crc32val, &ptr[nTrSamples * sizeof(int16_t)], sizeof crc32val); // idem
#if __BYTE_ORDER == __BIG_ENDIAN
      crc32val = __bswap_32(crc32val);
#endif
      return itsCrc32gen.checksum() == crc32val;
    }

    bool TBB_Dipole::hasAllZeroDataSamples(const TBB_Payload& payload, unsigned nTrSamples) const
    {
      /*
       * Good (noisy) data may have a few consecutive zero values, so this loop terminates
       * quickly, unless the antenna is broken or disabled, which happens sometimes.
       * Another reason for all zeros is that a wave generator is used and set to zero amp (pointless).
       * Unfortunately, the crc32 variant used does not reject all zeros (checksum 0).
       */
      for (unsigned i = 0; i < nTrSamples; i++) {
        if (payload.data[i] != 0) {
          return false;
        }
      }

      return true;
    }

  } // namespace Cobalt
} // namespace LOFAR

