//# TBB_Dipole.h: TBB per-dipole routines to store incoming TBB data
//# Copyright (C) 2012-2017  ASTRON (Netherlands Institute for Radio Astronomy)
//# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
//#
//# This file is part of the LOFAR software suite.
//# The LOFAR software suite is free software: you can redistribute it and/or
//# modify it under the terms of the GNU General Public License as published
//# by the Free Software Foundation, either version 3 of the License, or
//# (at your option) any later version.
//#
//# The LOFAR software suite is distributed in the hope that it will be useful,
//# but WITHOUT ANY WARRANTY; without even the implied warranty of
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//# GNU General Public License for more details.
//#
//# You should have received a copy of the GNU General Public License along
//# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
//#
//# $Id$

#ifndef LOFAR_COBALT_OUTPUTPROC_TBBDIPOLE_H
#define LOFAR_COBALT_OUTPUTPROC_TBBDIPOLE_H 1

#include "TBB_Frame.h"
#include <boost/scoped_ptr.hpp>
#include <boost/crc.hpp>

#include <Common/Thread/Mutex.h>
#include <Stream/FileStream.h>
#include <CoInterface/Parset.h>
#include <CoInterface/SmartPtr.h> // when switching to C++11, replace SmartPtr with unique_ptr

#include <dal/lofar/TBB_File.h>   // https://github.com/nextgen-astrodata/DAL

namespace LOFAR
{
  namespace Cobalt
  {

    class TBB_Dipole
    {
      /*
       * Note: Users don't want 1 raw file per dipole, and certainly not 1 raw file per subband (spectral mode).
       * But as long as LOFAR control doesn't provide dump duration or length in the parset, we cannot easily do any better.
       * In spectral mode, it is better for dispersed dumps and post-processing to store whole subbands (instead of spectra).
       * DAL doesn't support a subband nr in a DipoleDataset, so in the .h5 can only refer to 1 raw file...
       * Once LOFAR control supplies suitable parsets with ALL metadata, dumpInfo can be dumped...
       */
      boost::scoped_ptr<dal::TBB_DipoleDataset> itsDataset;

      struct DumpInfo {
        SmartPtr<LOFAR::FileStream> itsRawFile;
        std::vector<dal::Range> itsFlagOffsets;
        unsigned long long itsDatasetLen; // in values, including holes from missing data

        // These 2 fields are lazily initialized from the 1st received frame
        uint32_t itsTime0; // seconds
        uint32_t itsSampleNrOrSliceNr0; // transient mode: sampleNr; spectral mode: sliceNr
      };
      std::vector<DumpInfo> itsDumpInfo; // size: transient mode: 1, spectral mode: RSP_NR_SUBBANDS (indexed by band nr)

      time_t itsLastLogErrorTime;

      /*
       * Same truncated polynomial as standard crc32, but with:
       *   initial_remainder=0, final_xor_value=0, reflected_input=false, reflected_remainder_output=false
       * The boost::crc_optimal<> declarations precompute lookup tables,
       * so do not declare inside the checking routine. (Still, for every TBB_Dipole...)
       */
      boost::crc_optimal<32, 0x04C11DB7 /*, 0, 0, false, false*/> itsCrc32gen;

      // do not use
      TBB_Dipole& operator=(const TBB_Dipole& rhs);

    public:
      TBB_Dipole();
      TBB_Dipole(const TBB_Dipole& rhs); // do not use; only for TBB_Station std::vector<TBB_Dipole>(N) constr
      ~TBB_Dipole();

      // All TBB_Dipole objects are default constructed in a vector, so have init().
      void init(const TBB_Header& header, const Parset& parset, const StationMetaData& stationMetaData,
                const std::vector<double>& allSubbandCentralFreqs, const std::string& h5Filename,
                dal::TBB_Station& station, Mutex& h5Mutex);

      // Output threads
      bool isInitialized() const;

      void processTransientFrameData(const TBB_Frame& frame);
      void processSpectralFrameData(const TBB_Frame& frame);

    private:
      bool doTransient() const;
      uint32_t getTransientModeSampleNr(const TBB_Header& header) const;
      static std::string getRawFilename(const std::string& h5Filename, bool doTransient,
                                        unsigned rspID, unsigned rcuID);
      void appendFlags(DumpInfo& di, size_t offset, size_t len);
      // initTBB_DipoleDataset() must be called with the global h5Mutex held.
      void initTBB_DipoleDataset(const TBB_Header& header, const Parset& parset,
                                 const StationMetaData& stationMetaData,
                                 const std::vector<double>& allSubbandCentralFreqs,
                                 const std::string& rawFilename, dal::TBB_Station& station);
      bool crc32tbb(const TBB_Payload* payload, size_t nTrSamples);
      bool hasAllZeroDataSamples(const TBB_Payload& payload, unsigned nTrSamples) const;
    };

  } // namespace Cobalt
} // namespace LOFAR

#endif

